from typing import ClassVar

from django.db.models import CharField, Count, Index, Manager, Model, PositiveSmallIntegerField


class FakeDataManager(Manager):
    def provider_count(self, provider):
        return self.filter(provider=provider).values("provider").annotate(count=Count("provider")).values("count")


class FakeData(Model):
    provider = CharField(max_length=255, verbose_name="Faker provider", db_index=True)
    provider_offset = PositiveSmallIntegerField()
    content = CharField(max_length=255, verbose_name="Fake content")

    objects = FakeDataManager()

    class Meta:
        indexes: ClassVar[list[Index]] = [
            Index(fields=["provider", "provider_offset"]),
        ]
        unique_together = (("provider", "provider_offset"),)

    def __str__(self):
        return f"{self.provider}: '{self.content}'"



import sys

from django.core.management.base import BaseCommand

from django_scrubber.services.validator import ScrubberValidatorService


class Command(BaseCommand):
    def handle(self, *args, **options):
        service = ScrubberValidatorService()
        non_scrubbed_field_list = service.process()

        found_models = 0
        found_fields = 0

        if len(non_scrubbed_field_list):
            for model_path, affected_field_list in non_scrubbed_field_list.items():
                found_models += 1
                for _field in affected_field_list:
                    found_fields += 1

            if found_models > 0:
                sys.exit(1)



import importlib
import logging
import warnings

from django.apps import apps
from django.conf import settings
from django.contrib.sessions.models import Session
from django.core.exceptions import FieldDoesNotExist
from django.core.management.base import BaseCommand, CommandError
from django.db.utils import DataError, IntegrityError

from django_scrubber import settings_with_fallback
from django_scrubber.models import FakeData
from django_scrubber.scrubbers import Keep
from django_scrubber.services.validator import ScrubberValidatorService

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Replace database data according to model-specific or global scrubbing rules."
    leave_locale_alone = True

    def add_arguments(self, parser):
        parser.add_argument(
            "--model",
            type=str,
            required=False,
            help="Scrub only a single model (format <app_label>.<model_name>)",
        )
        parser.add_argument(
            "--keep-sessions",
            action="store_true",
            required=False,
            help="Will NOT truncate all (by definition critical) session data",
        )
        parser.add_argument(
            "--remove-fake-data",
            action="store_true",
            required=False,
            help="Will truncate the database table storing preprocessed data for the Faker library. "
            "If you want to do multiple iterations of scrubbing, it will save you time to keep "
            "them. If not, you will add a huge bunch of data to your dump size.",
        )

    def handle(self, *args, **kwargs):
        if not settings.DEBUG:
            # avoid logger, otherwise we might silently fail if we're on live and logging is being sent somewhere else
            self.stderr.write("This command should only be run with DEBUG=True, to avoid running on live systems")
            return False

        # Check STRICT mode
        if settings_with_fallback("SCRUBBER_STRICT_MODE"):
            service = ScrubberValidatorService()
            non_scrubbed_field_list = service.process()
            if len(non_scrubbed_field_list) > 0:
                self.stderr.write(
                    'When "SCRUBBER_STRICT_MODE" is enabled, you have to define a scrubbing policy '
                    "for every text-based field.",
                )
                return False

        global_scrubbers = settings_with_fallback("SCRUBBER_GLOBAL_SCRUBBERS")

        # run only for selected model
        if kwargs.get("model") is not None:
            app_label, model_name = kwargs.get("model").rsplit(".", 1)
            try:
                models = [apps.get_model(app_label=app_label, model_name=model_name)]
            except LookupError as e:
                raise CommandError("--model should be defined as <app_label>.<model_name>") from e

        # run for all models of all apps
        else:
            models = apps.get_models()

        scrubber_apps_list = settings_with_fallback("SCRUBBER_APPS_LIST")
        for model_class in models:
            self._scrub_model(model_class, scrubber_apps_list, global_scrubbers)

        # Truncate session data
        if not kwargs.get("keep_sessions", False):
            Session.objects.all().delete()

        # Truncate Faker data
        if kwargs.get("remove_fake_data", False):
            FakeData.objects.all().delete()
            return None
        return None

    def _scrub_model(self, model_class, scrubber_apps_list, global_scrubbers):
        if model_class._meta.proxy:
            return
        if settings_with_fallback("SCRUBBER_SKIP_UNMANAGED") and not model_class._meta.managed:
            return
        if scrubber_apps_list and model_class._meta.app_config.name not in scrubber_apps_list:
            return

        scrubbers = {}
        for field in model_class._meta.fields:
            if field.name in global_scrubbers:
                scrubbers[field] = global_scrubbers[field.name]
            elif type(field) in global_scrubbers:
                scrubbers[field] = global_scrubbers[type(field)]

        scrubbers.update(_get_model_scrubbers(model_class))

        # Filter out all fields marked as "to be kept"
        scrubbers_without_kept_fields = {}
        for field, scrubbing_method in scrubbers.items():
            if scrubbing_method != Keep:
                scrubbers_without_kept_fields[field] = scrubbing_method
        scrubbers = scrubbers_without_kept_fields

        if not scrubbers:
            return

        realized_scrubbers = _filter_out_disabled(_call_callables(scrubbers))

        logger.info("Scrubbing %s with %s", model_class._meta.label, realized_scrubbers)

        try:
            model_class.objects.update(**realized_scrubbers)
        except IntegrityError as e:
            raise CommandError(
                f"Integrity error while scrubbing {model_class} ({e}); maybe increase SCRUBBER_ENTRIES_PER_PROVIDER?",
            ) from e
        except DataError as e:
            raise CommandError(f"DataError while scrubbing {model_class} ({e})") from e


def _call_callables(d):
    """
    Helper to realize lazy scrubbers, like Faker, or global field-type scrubbers
    """
    return {k.name: ((callable(v) and v(k)) or v) for k, v in d.items()}


def _parse_scrubber_class_from_string(path: str):
    """
    Takes a string to a certain scrubber class and returns a python class definition - not an instance.
    """
    try:
        module_name, class_name = path.rsplit(".", 1)
        module = importlib.import_module(module_name)
        return getattr(module, class_name)
    except (ImportError, ValueError) as e:
        raise ImportError(f'Mapped scrubber class "{path}" could not be found.') from e


def _get_model_scrubbers(model):
    # Get model-scrubber-mapping from settings
    scrubber_mapping = settings_with_fallback("SCRUBBER_MAPPING")

    # Initialise scrubber list
    scrubbers = {}

    # Check if model has a settings-defined...
    if model._meta.label in scrubber_mapping:
        scrubber_cls = _parse_scrubber_class_from_string(scrubber_mapping[model._meta.label])
    # If not...
    else:
        # Try to get the scrubber metaclass from the given model
        try:
            scrubber_cls = model.Scrubbers
        except AttributeError:
            return scrubbers  # no model-specific scrubbers

    # Get field mappings from scrubber class
    for k, v in _get_fields(scrubber_cls):
        try:
            field = model._meta.get_field(k)
            scrubbers[field] = v
        except FieldDoesNotExist:
            warnings.warn(f"Scrubber defined for {model.__name__}.{k} but field does not exist", stacklevel=2)

    # Return scrubber-field-mapping
    return scrubbers


def _get_fields(d):
    """
    Helper to get "normal" (i.e.: non-magic and non-dunder) instance attributes.
    Returns an iterator of (field_name, field) tuples.
    """
    return ((k, v) for k, v in vars(d).items() if not k.startswith("_"))


def _filter_out_disabled(d):
    """
    Helper to remove Nones (actually any false-like type) from the scrubbers.
    This is needed so we can disable global scrubbers in a per-model basis.
    """
    return {k: v for k, v in d.items() if v}



#!/usr/bin/env python
import re
from pathlib import Path

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version():
    """Retrieves the version from django_scrubber/__init__.py"""
    with Path("django_scrubber/__init__.py").open("r") as f:
        version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


version = get_version()

with Path("README.md").open(mode="r") as f:
    readme = f.read()
with Path("CHANGELOG.md").open(mode="r") as f:
    changelog = f.read()

setup(
    name="django-scrubber",
    version=version,
    description="""Data Anonymizer for Django""",
    long_description=readme + "\n\n" + changelog,
    long_description_content_type="text/markdown",
    author="RegioHelden GmbH",
    author_email="entwicklung@regiohelden.de",
    url="https://github.com/regiohelden/django-scrubber",
    project_urls={
        "Documentation": "https://github.com/RegioHelden/django-scrubber/blob/master/README.md",
        "Maintained by": "https://github.com/RegioHelden/django-scrubber/blob/master/AUTHORS.md",
        "Bugtracker": "https://github.com/RegioHelden/django-scrubber/issues",
        "Changelog": "https://github.com/RegioHelden/django-scrubber/blob/master/CHANGELOG.md",
    },
    packages=[
        "django_scrubber",
    ],
    include_package_data=True,
    install_requires=[
        "faker>=20.0.0",
    ],
    license="BSD",
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.0",
        "Framework :: Django :: 5.1",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Security",
        "Topic :: Software Development",
    ],
)



from django.conf import settings
from django.db import models

__version__ = "4.0.0"

defaults = {
    "SCRUBBER_RANDOM_SEED": 42,  # we prefer idempotent scrubbing
    "SCRUBBER_ENTRIES_PER_PROVIDER": 1000,
    "SCRUBBER_GLOBAL_SCRUBBERS": {},
    "SCRUBBER_SKIP_UNMANAGED": True,
    "SCRUBBER_APPS_LIST": None,
    "SCRUBBER_ADDITIONAL_FAKER_PROVIDERS": {*()},
    "SCRUBBER_FAKER_LOCALE": None,
    "SCRUBBER_MAPPING": {},
    "SCRUBBER_STRICT_MODE": False,
    "SCRUBBER_REQUIRED_FIELD_TYPES": (
        models.CharField,
        models.TextField,
        models.URLField,
        models.JSONField,
        models.GenericIPAddressField,
        models.EmailField,
    ),
    "SCRUBBER_REQUIRED_FIELD_MODEL_WHITELIST": (
        "auth.Group",
        "auth.Permission",
        "contenttypes.ContentType",
        "db.TestModel",
        "sessions.Session",
        "sites.Site",
        "django_scrubber.FakeData",
    ),
}


def settings_with_fallback(key):
    return getattr(settings, key, defaults[key])


class ScrubberInitError(Exception):
    pass
